import pandas as pd
import numpy as np
import logging
import os
from nemosis import dynamic_data_compiler, static_table
from nemed import *
from nemed.downloader import download_unit_dispatch
from datetime import datetime as dt, timedelta
from nemglo.backend import input_validation as inv

logger = logging.getLogger(__name__)

class Market:
    def __init__(self, local_cache=r'.\cache', intlength=5, region=None, start_date=None, end_date=None, duid_1=None, \
        duid_2=None, emissions_type=None):
        logger.info("Establish nemosis_data logger")
        self._check_data_inputs(local_cache, intlength, region, start_date, end_date, duid_1, duid_2, emissions_type)
        local_cache = _check_cache(local_cache)
        # Inputs
        self._cache = local_cache
        self._intlength = intlength
        self._region = region
        self._start = start_date + ":00"
        self._end = end_date + ":00"
        self._duid_1 = duid_1
        self._duid_2 = duid_2
        self._emissions_type = emissions_type
        # Outputs
        self._info = None
        self._prices = None
        self._generation = None

    def _check_data_inputs(self, local_cache, intlength, region, start_date, end_date, duid_1, duid_2, \
        emissions_type):
        """Validate correct argument types have been used.
        """
        # Required arguments
        inv.validate_variable_type(intlength, int, "intlength", "market_data")
        inv.validate_variable_type(local_cache, str, "local_cache", "market_data")
        inv.validate_variable_type(region, str, "region", "market_data")
        inv.validate_variable_type(start_date, str, "start_date", "market_data")
        inv.validate_variable_type(end_date, str, "end_date", "market_data")
        inv.validate_and_convert_date(start_date, "start_date")
        inv.validate_and_convert_date(end_date, "start_date")

        # Allowed arguments
        if duid_1 != None:
            inv.validate_variable_type(duid_1, str, "duid_1", "market_data")
        if duid_2 != None:
            inv.validate_variable_type(duid_2, str, "duid_2", "market_data")
        if emissions_type != None:
            inv.validate_variable_type(emissions_type, str, "emissions_type", "market_data")
            _check_statement(((emissions_type == 'Total') | (emissions_type == 'Marginal')), "Invalid `emissions_type` \
                input. Must be either of ['Total','Marginal']")
    
    def get_prices(self):
        """
        Retrieve the NEM historical regional reference node price for the 
        defined parameters of the nemosis data loader object.

        Returns
        -------
        pd.DataFrame()
            Returns pd.DataFrame with columns = ['Time','Prices'].
        """
        prices = dynamic_data_compiler(start_time=self._start, 
                                       end_time=self._end, 
                                       table_name='DISPATCHPRICE', 
                                       raw_data_location=self._cache,
                                       fformat='feather',
                                       filter_cols=['REGIONID'],
                                       filter_values=([[self._region]]),
                                       select_columns=['SETTLEMENTDATE','REGIONID','RRP'],
                                       keep_csv=False)
        prices.set_index('SETTLEMENTDATE',inplace=True)

        # Resample data if requested interval length is not 5 minutes
        if self._intlength != 5:
            self._prices = pd.DataFrame(prices.resample(str(self._intlength)+'min', 
                                  label='right', origin='end').mean())
        else:
            self._prices = pd.DataFrame(prices[['RRP']])

        self._prices.reset_index(inplace=True)
        self._prices.columns = ['Time', 'Prices']
        return self._prices


    def get_generation(self, cf_trace=True):
        df = self._download_all_unitdispatch()
        if not cf_trace:
            return df
        else:
            traces = self._calculate_capacity_factor_traces(df, region_filter=True)
            traces[traces < 0] = 0
            self._generation = traces.reset_index()
            if (self._duid_1 != None) & (self._duid_2 != None):
                return self._generation[['Time', self._duid_1, self._duid_2]]
            elif (self._duid_1 != None):
                return self._generation[['Time', self._duid_1]]
            elif (self._duid_2 != None):
                return self._generation[['Time', self._duid_2]]
            else:
                return None


    def _download_geninfo(self, filter_regions=['NSW1','QLD1','VIC1','SA1','TAS1'], filter_tech=None):
        """
        Downloads the Generator Registration and Exemption List data from AEMO.
        """

        gen_info = static_table(table_name="Generators and Scheduled Loads",
                          raw_data_location=self._cache,
                          select_columns=['Station Name','DUID','Region',
                          'Fuel Source - Descriptor','Reg Cap (MW)'],
                          filter_cols=['Region'],
                          filter_values=[filter_regions])
        print(gen_info)
        gen_info['Reg Cap (MW)'] = np.where(gen_info['Reg Cap (MW)'] == '-', np.nan, \
            gen_info['Reg Cap (MW)'])
        gen_info['Reg Cap (MW)'] = gen_info['Reg Cap (MW)'].astype(float)

        if filter_tech:
            self._info = gen_info[(gen_info['Reg Cap (MW)'].astype(float) >= 30) & \
            (gen_info['Fuel Source - Descriptor'].isin(filter_tech))]
        else:
            self._info = gen_info[(gen_info['Reg Cap (MW)'].astype(float) >= 30) & \
                (gen_info['Fuel Source - Descriptor'].isin(['Wind','Solar','solar']))]


    def _calculate_capacity_factor_traces(self, dataset, region_filter):
        """
        Divides the DUID generation dispatch values by the registered capacity
        for the DUID as found in AEMO's Registration and Exemption List
        """
        if region_filter:
            self._download_geninfo(filter_regions=[self._region])
        else:
            self._download_geninfo()
        info = self._info
        result = dataset.copy()

        for unit in result.columns:
            regcap = info[info['DUID']==unit]['Reg Cap (MW)']
            if not regcap.empty:
                if isinstance(regcap.values[0],float):
                    result[unit] = result[unit].div(float(regcap))
                else:
                    result.drop(unit, axis=1, inplace=True)
            else:
                result.drop(unit, axis=1, inplace=True)
        result.columns.name = None
        return result

    def _download_all_unitdispatch(self):

        # Download via nemed function
        dispatch_table = self._download_unit_dispatch(start_time=self._start,
                                                      end_time=self._end,
                                                      cache=self._cache,
                                                      return_all=False,
                                                      check=True)
        # Set DUIDs as columns
        # (aggfunc purpose is to resolve mismatching Scada and InitialMW)
        vrep = dispatch_table.pivot(index='Time', values='Dispatch', columns='DUID')

        # Resample data if requested interval length is not 5 minutes
        if self._intlength != 5:
            result = vrep.resample(str(self._intlength)+'min', 
                                label='right', origin='end').mean()
        else:
            result = vrep
        return result


    def _download_unit_dispatch(self, start_time, end_time, cache, return_all=True, check=True):
        """Downloads historical SCADA dispatch data via NEMOSIS. Inherited from NEMED

        Parameters
        ----------
        start_time : str
            Start Time Period in format 'yyyy/mm/dd HH:MM:SS'
        end_time : str
            End Time Period in format 'yyyy/mm/dd HH:MM:SS'
        cache : str
            Raw data location in local directory
        source_initialmw : bool
            Whether to download initialmw column from DISPATCHLOAD table, by default False
        source_scada : bool
            Whether to download scada column from DISPATCH_UNIT_SCADA table, by default True
        overwrite : str
            The data value to overwrite in the returned 'Dispatch' column if there is a discrepency in initialmw and scada.
            Must be one of ['initialmw','scada','average']. If one of source_initialmw or source_scada is False, `overwrite`
            has null effect. By default 'scada'.
        return_all : bool
            Whether to return all columns or only ['Time','DUID','Dispatch'], by default False.
        check : bool
            Whether to check for, and remove duplicates after function is complete, by default True.

        Returns
        -------
        pd.DataFrame
            Returns generation data as per NEMOSIS

        """
        DISPATCH_INT_MIN = 5
        # Check inputs
        cache = hp._check_cache(cache)
        assert(isinstance(start_time, str)), "`start_time` must be a string in format yyyy/mm/dd HH:MM:SS"
        assert(isinstance(end_time, str)), "`end_time` must be a string in format yyyy/mm/dd HH:MM:SS"

        # Adjust timestamps for Scada interval-beginning
        shift_stime = dt.strptime(start_time, "%Y/%m/%d %H:%M:%S")
        shift_stime = shift_stime + timedelta(minutes=DISPATCH_INT_MIN)
        shift_etime = dt.strptime(end_time, "%Y/%m/%d %H:%M:%S")
        shift_etime = shift_etime + timedelta(minutes=DISPATCH_INT_MIN)
        get_start_time = dt.strftime(shift_stime, "%Y/%m/%d %H:%M:%S")
        get_end_time = dt.strftime(shift_etime, "%Y/%m/%d %H:%M:%S")

        # Download Dispatch Unit Scada table via NEMOSIS (this includes Non-Scheduled generators)
        disp_scada = dynamic_data_compiler(
            start_time=get_start_time,
            end_time=get_end_time,
            table_name="DISPATCH_UNIT_SCADA",
            raw_data_location=cache,
            select_columns=["SETTLEMENTDATE", "DUID", "SCADAVALUE"],
            fformat="feather",
        )
        disp_scada["Time"] = disp_scada["SETTLEMENTDATE"] - timedelta(minutes=DISPATCH_INT_MIN)
        master = disp_scada[['Time', 'DUID', 'SCADAVALUE']]
        master['Dispatch'] = master['SCADAVALUE']

        # Final check for intervention periods and duplicates entries
        if check:
            final = self._clean_duplicates(master)
        else:
            final = master

        # Return dataset
        if return_all:
            return final
        else:
            return final[['Time', 'DUID', 'Dispatch']]


    def _clean_duplicates(self, table, value_col="Dispatch"):
        """Filtering inherited from NEMED process
        """
        if any(table.duplicated(subset=['Time', 'DUID'])):
            print("Duplicate Timestamped DUIDs found. Updating dataset for duplicates.")
            # Take average values where duplicates differ
            table_clean = table.pivot_table(index=["Time", "DUID"], values=value_col, aggfunc=np.mean)
            table_clean = table_clean.reset_index()

            # Remove duplicates where Time and DUID match
            table_clean = table_clean.drop_duplicates(subset=["Time", "DUID"])
            return table_clean
        else:
            return table


    def estimate_ppa_costs(self, vre_capacity, strike_price, floor_price):
        ## NEED TO ADD INPUT and PREPROCESSES COMPLETE CHECKS

        # always using duid_1

        cfd_volume = self._generation[self._duid_1] * vre_capacity

        if floor_price != None:
            cfd_value = [(strike_price - max(floor_price, self._prices['Prices'][i])) * (self._intlength/60) \
                for i in range(0, len(self._prices['Prices']))]
        else:
             cfd_value = [(strike_price - self._prices['Prices'][i]) * (self._intlength/60) \
                for i in range(0, len(self._prices['Prices']))]           

        cfd = pd.DataFrame({'Time': [self._prices['Time'][i] for i in range(0, min(len(cfd_value), len(cfd_volume)))],
                            'Cost': [cfd_volume[i] *cfd_value[i] for i in range(0, min(len(cfd_value), len(cfd_volume)))],
                            'PriceDiff': [cfd_value[i] for i in range(0, min(len(cfd_value), len(cfd_volume)))],
                            'Volume': [cfd_volume[i] for i in range(0, min(len(cfd_value), len(cfd_volume)))]})
        return cfd


    def get_emissions_intensity(self):
        """_summary_

        USING MARGINAL EMISSIONS: Error may occur with SSL and requests module. See fix: https://www.youtube.com/watch?v=mN8SLBsvSCU

        Returns
        -------
        _type_
            _description_
        """
        emissions_type = self._emissions_type
        if emissions_type == "Total":

            nemed_df = get_total_emissions(start_time=self._start[:-3],
                                           end_time=self._end[:-3],
                                           cache=self._cache,
                                           filter_regions=[self._region],
                                           by=None,
                                           generation_sent_out=True,
                                           assume_energy_ramp=True,
                                           return_pivot=False)
            nemed_df.rename(columns={'TimeEnding': 'Time'},inplace=True)

            # Adjust time resolution to set interval length
            if self._intlength != 5:
                emissions = nemed_df.set_index('Time').resample(str(self._intlength)+'min', label='right', origin='end').sum()
                emissions['Intensity_Index'] = emissions['Total_Emissions'] / emissions['Energy']
                emissions.reset_index(inplace=True)
            else:
                emissions = nemed_df
            return emissions[['Time', 'Intensity_Index']]

        elif emissions_type == "Marginal":
            
            nemed_df = get_marginal_emissions(start_time=self._start[:-3],
                                              end_time=self._end[:-3],
                                              cache=self._cache+"_2")

            nemed_df = nemed_df[nemed_df['Region'] == self._region]

            # Adjust time resolution to set interval length
            if self._intlength != 5:
                emissions = nemed_df.set_index('Time').resample(str(self._intlength)+'min', label='right', origin='end').mean()
                emissions.reset_index(inplace=True)
            else:
                emissions = nemed_df
            return emissions[['Time', 'Intensity_Index']]           


def _check_statement(statement, err_message, log_type=logger.exception):
    try:
        assert statement, err_message
    except AssertionError as e:
        log_type(e)
        raise e


def convert_dt_order(string):
    """
    Convert defined input datetime string into format required by nemosis

    Parameters
    ----------
    string : str
        Input datetime string as ('%d/%m/%Y %H:%M').

    Returns
    -------
    ret_str : str
        Output datetime string as ('%Y/%m/%d %H:%M:%S').
    """
    ret_str = dt.strptime(string,'%d/%m/%Y %H:%M')
    ret_str = dt.strftime(ret_str, '%Y/%m/%d %H:%M:%S')
    return ret_str

def _check_cache(cache):
    """Check the cache folder directory exists
    Parameters
    ----------
    cache : str
        Folder path location to be used as local cache.
    Returns
    -------
    str
        Existing (or updated if error) cache directory path.
    """
    try:
        assert(isinstance(cache, str))
    except AssertionError as e:
        logger.exception("Cache input must be a string")
        raise e

    if not os.path.isdir(cache):
        cache = os.path.join(os.getcwd(), "cache_nemglo")
        if os.path.isdir(cache):
            logger.warning(f"Input cache path is invalid or not found. Using existing cache directory at {cache}")
        else:
            os.mkdir(cache)
            logger.warning(f"Input cache path is invalid or not found. Creating new cache directory at {cache}")
    return cache


# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['nemglo', 'nemglo.backend', 'nemglo.components', 'nemglo.planning']

package_data = \
{'': ['*']}

install_requires = \
['flask-cors>=3.0.10,<4.0.0',
 'flask>=2.2.2,<3.0.0',
 'mip>=1.13.0,<2.0.0',
 'nemed',
 'nemosis>=3.1.0,<4.0.0',
 'numpy>=1.23.1,<2.0.0',
 'pandas>=1.4.3,<2.0.0',
 'pyyaml>=6.0,<7.0']

setup_kwargs = {
    'name': 'nemglo',
    'version': '0.2.2',
    'description': 'Green-energy Load Optimisation tool for the NEM',
    'long_description': '# NEMGLO\nNEMGLO, the Green-energy Load Optimisation tool for the Australian National Electricity Market (NEM) is a python tool allows users to counterfactually assess flexible load operating strategies in the NEM using historical market data. Specifically, the tool enables users to test various scenarios with differing PPA contracting structures and green-certificate schemes, a primary case study being the ability to \'validate\' the integrity of "green"-hydrogen produced via grid-connected electrolysis with bundled PPA contracting. Although the tool was developed contextually for electrolyser loads, the functionalities can inherently be abstracted to commercial/industrial flexible loads seeking to consume green-energy. \n\n**Customisable Tool Features:**\n- Load Operating Characteristics (Min stable load, ramp rates, etc.)\n- Multiple Power Purchase Agreements (contract volume, strike, floor, etc.)\n\n**Features Coming Soon:**\n- Renewable Energy Certificate procurement + surrender (with/without temporal matching)\n- Shadow Pricing of Grid Emissions Intensity (average & marginal emissions)\n- Constrain to Green Energy Certification Standards for H2 (max tCO2 content per tH2) \n\n**Project Links**\n- Project Homepage: https://www.nemglo.org/\n- Python Github Repository: https://github.com/dec-heim/NEMGLO\n- Python Documentation: https://nemglo.readthedocs.io/en/latest/about.html\n- NEMGLO App Github Repository: https://github.com/dec-heim/NEMGLO-app\n\n\n## Installation\n```bash\npip install nemglo\n```\nFor guidance on installation and usage of [NEMGLO-app](https://github.com/dec-heim/NEMGLO-app) (a graphical user interface), please refer to its project page.\n\n## Usage\nFor guidance on `NEMGLO` usage, see the [Examples]() section of the documentation.\n\n## Contributing\nInterested in contributing? Check out the [contributing guidelines](CONTRIBUTING.md).\nPlease note that this project is released with a [Code of Conduct](). By contributing to this project, you agree to abide by its terms.\n\n## License\n`NEMGLO` was created by Declan Heim as a Master\'s Project at UNSW. It is licensed under the terms of the [BSD 3-Clause license](LICENSE).\n\n## Credits\nThe `NEMGLO` project is affilitated with the [UNSW Collaboration on Energy and Environmental Markets](https://www.ceem.unsw.edu.au/). The [`NEMGLO-app`](https://github.com/dec-heim/NEMGLO-app) was financially supported by the [UNSW Digital Grid Futures Institute](https://www.dgfi.unsw.edu.au/) to Dec-2022.\n\n`NEMGLO` incorporates functionality from a suite of UNSW-CEEM tools, namely, [`NEMOSIS`](https://github.com/UNSW-CEEM/NEMOSIS) to extract historical market data and [`NEMED`](https://github.com/UNSW-CEEM/NEMED) to compute emissions data from AEMO\'s MMS databases respectively. The structure of the optimiser codebase is further adopted from sister tool [`nempy`](https://github.com/UNSW-CEEM/nempy) under [nempy licence](https://github.com/UNSW-CEEM/nempy/blob/master/LICENSE).\n\n### Acknowlgements\n\nMany thanks to:\n- Jay Anand, co-developer of `NEMGLO-app`. \n- Nick Gorman, Abhijith Prakash, and Shayan Naderi for pointers on `NEMGLO` code development (including adopted NEM tools).\n- Iain MacGill, Anna Bruce, Rahman Daiyan, and Jack Sheppard for advice at various stages to this project.\n\n## Contact\nQuestions and feedback are very much welcomed. Please reach out by email to [declanheim@outlook.com](mailto:declanheim@outlook.com)',
    'author': 'dec-heim',
    'author_email': '92137442+dec-heim@users.noreply.github.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



# class DataFrameSchema and class SeriesSchema creditted to N.Gorman (nempy)
import pandas as pd
import numpy as np
from datetime import datetime as dt
import logging

logger = logging.getLogger(__name__)

class DataFrameSchema:
    def __init__(self, name, primary_keys=None):
        self.name = name
        self.primary_keys = primary_keys
        self.columns = {}
        self.required_columns = []

    def add_column(self, column, optional=False):
        self.columns[column.name] = column
        if not optional:
            self.required_columns.append(column.name)

    def validate(self, df):
        for col in df:
            if col not in self.columns:
                raise UnexpectedColumn("Column '{}' is not allowed in DataFrame '{}'. Permitted columns are {}"\
                    .format(col, self.name, list(self.columns.keys())))

        for col in self.required_columns:
            if col not in df.columns:
                raise MissingColumnError("Column '{}' not in DataFrame '{}'. Required columns are {}"\
                    .format(col, self.name, self.required_columns))

        for col in self.columns:
            if col in df.columns:
                self.columns[col].validate(df[col])

        if self.primary_keys is not None:
            self._check_for_repeated_rows(df)

    def _check_for_repeated_rows(self, df):
        cols_in_df = [col for col in self.primary_keys if col in df.columns]
        if len(df.index) != len(df.drop_duplicates(cols_in_df)):
            raise RepeatedRowError('{} should only have one row for each {}.'.format(self.name, ' '.join(cols_in_df)))


class SeriesSchema:
    def __init__(self, name, data_type, allowed_values=None, must_be_real_number=False, not_negative=False,
                 no_duplicates=False, ascending_order=False, minimum=None, maximum=None):
        self.name = name
        self.data_type = data_type
        self.allowed_values = allowed_values
        self.must_be_real_number = must_be_real_number
        self.not_negative = not_negative
        self.no_duplicates = no_duplicates
        self.ascending_order = ascending_order
        self.min = minimum
        self.max = maximum

    def validate(self, series):
        self._check_data_type(series)
        self._check_allowed_values(series)
        self._check_is_real_number(series)
        self._check_is_not_negtaive(series)
        self._check_no_duplicates(series)
        self._check_is_ordered(series)
        self._check_min_max(series)

    def _check_data_type(self, series):
        if self.data_type == str:
            if not all(series.apply(lambda x: type(x) == str)):
                raise ColumnDataTypeError("All elements of column '{}' should have type str".format(self.name))
        elif self.data_type == callable:
            if not all(series.apply(lambda x: callable(x))):
                raise ColumnDataTypeError("All elements of column '{}' should have type callable".format(self.name))
        elif self.data_type != series.dtype:
            raise ColumnDataTypeError("Column '{}' should have type '{}'".format(self.name, self.data_type))

    def _check_allowed_values(self, series):
        if self.allowed_values is not None:
            if not series.isin(self.allowed_values).all():
                raise ColumnValues("The column '{}' can only contain the values '{}'.".format(self.name, \
                    self.allowed_values))

    def _check_is_real_number(self, series):
        if self.must_be_real_number:
            if np.inf in series.values:
                raise ColumnValues("Value inf not allowed in column '{}'.".format(self.name))
            if np.NINF in series.values:
                raise ColumnValues("Value -inf not allowed in column '{}'.".format(self.name))
            if series.isnull().any():
                raise ColumnValues("Null values not allowed in column '{}'.".format(self.name))

    def _check_is_not_negtaive(self, series):
        if self.not_negative:
            if series.min() < 0.0:
                raise ColumnValues("Negative values not allowed in column '{}'.".format(self.name))

    def _check_no_duplicates(self, series):
        if self.no_duplicates:
            if series.duplicated().any():
                raise ColumnValues("Duplicate values not allowed in column '{}'.".format(self.name))

    def _check_is_ordered(self, series):
        if self.ascending_order:
            if not all(series.reset_index(drop=True) == series.sort_values(ascending=True).reset_index(drop=True)):
                raise ColumnValues("Values must be sequenced in ascending order in column '{}'.".format(self.name))

    def _check_min_max(self, series):
        if self.min:
            if series.min() < self.min:
                raise ColumnValues("Minimum allowed value in column '{}' is {}.".format(self.name, self.min))
        if self.max:
            if series.max() > self.max:
                raise ColumnValues("Maximum allowed value in column '{}' is {}.".format(self.name, self.max))


def validate_unique_id(c_definition, system_plan, identifier):
    for other_gen in [c for c in system_plan._components if c.__class__ is c_definition]:
        if other_gen._id == identifier:
            raise ComponentError("Cannot create a {} object of the same name as an existing instance in Plan '{}'."\
                .format(c_definition, system_plan._id))


def validate_existing(system_plan, name, as_var=False, as_objective_cost=False, as_constr_lhs=False, \
    as_constr_rhs=False, as_constr_rhs_dynamic=False, as_constr_bigM=False, as_sos_2=False):
    if as_var:
        if name in system_plan._var:
            raise ComponentError("Cannot create a new variable {} as it already exists in Plan '{}'."\
                .format(name, system_plan._id))
    if as_objective_cost:
        if name in system_plan._objective_cost:
            raise ComponentError("Cannot create a new objective cost {} as it already exists in Plan '{}'."\
                .format(name, system_plan._id))
    if as_constr_lhs:
        if name in system_plan._constr_lhs:
            raise ComponentError("Cannot create a new constraint LHS {} as it already exists in Plan '{}'."\
                .format(name, system_plan._id))
    if as_constr_rhs:
        if name in system_plan._constr_rhs:
            raise ComponentError("Cannot create a new constraint RHS {} as it already exists in Plan '{}'."\
                .format(name, system_plan._id))
    if as_constr_rhs_dynamic:
        if name in system_plan._constr_rhs_dynamic:
            raise ComponentError("Cannot create a new constraint RHS dynamic {} as it already exists in Plan '{}'."\
                .format(name, system_plan._id))
    if as_constr_bigM:
        if name in system_plan._constr_bigM:
            raise ComponentError("Cannot create a new constraint bigM {} as it already exists in Plan '{}'."\
                .format(name, system_plan._id))
    if as_sos_2:
        if name in system_plan._sos_2:
            raise ComponentError("Cannot create a new SOS-2 {} as it already exists in Plan '{}'."\
                .format(name, system_plan._id))
  

def validate_positive_float(input, in_name, c_name):
    assert isinstance(input, float), "{} Argument: '{}' must be a float.".format(c_name, in_name)
    assert input >= 0, "{} Argument: '{}' must be a positive number.".format(c_name, in_name)

def validate_variable_type(var, vartype, inputname, functionname):
    try:
        assert isinstance(var, vartype), f"`{inputname}` of `{functionname}` cannot be {var}. Accepted types: {vartype}"
    except AssertionError as e:
        logger.exception(e)
        raise e

def validate_and_convert_date(date_string, inputname, date_format="%Y/%m/%d %H:%M"):
    try:
        date_obj = dt.strptime(date_string, date_format)
        return date_obj
    except ValueError:
        e = (f"Invalid date string was passed {date_string} as {inputname}. Required date format: {date_format}")
        logger.exception(e)


def validate_get_market_data(conf):
    """Validate required and allowed fields for '/api/get-market-data' call.
    """
    _check_fields(d=conf,
                  name="conf",
                  required={"market_data"},
                  allowed={"emissions_data"})

    _check_fields(d=conf['market_data'],
                  name="conf['market_data']",
                  required={"start_date", "end_date", "region", "dispatch_interval_length"},
                  allowed={"start_time","end_time"}
                  )

    if 'emissions_data' in conf:
        _check_fields(d=conf['emissions_data'],
                      name="conf['emissions_data']",
                      required={"emissions_type"})


def validate_get_generator_data(conf):
    """Validate required and allowed fields for '/api/get-generator-data' call.    
    """
    _check_fields(d=conf,
                  name="conf",
                  required={"market_data","ppa"},
                  allowed={""})

    _check_fields(d=conf['market_data'],
                  name="conf['market_data']",
                  required={"start_date", "end_date", "region", "dispatch_interval_length"},
                  allowed={"start_time","end_time"}
                  )

    _check_fields(d=conf['ppa'],
                  name="conf['ppa']",
                  required={"duid", "capacity", "strike_price"},
                  allowed={"floor_price"})


def validate_get_data(conf):
    """Validate required and allowed fields for '/api/get-data' call.    
    """
    _check_fields(d=conf,
                  name="conf",
                  required={"market_data", "electrolyser_load", "ppa_1", "ppa_2", "rec"},
                  allowed={""})
    logger.warning("No subfield checks on '/api/get-data'")
    # _check_fields(d=conf['market_data'],
    #               name="conf['market_data']",
    #               required={"start_date", "end_date", "region", "dispatch_interval_length"},
    #               allowed={""}
    #               )

    # _check_fields(d=conf['ppa'],
    #               name="conf['ppa']",
    #               required={"duid", "capacity", "strike_price"},
    #               allowed={"floor_price"})


def _check_fields(d, name="", required={""}, allowed={""}, log_type=logger.exception):
    err_message = f"API param: {name} has invalid fields. Required fields are: {required}. Allowed fields are: {allowed}"
    try:
        assert required <= d.keys() <= (required | allowed), err_message
    except AssertionError as e:
        log_type(e)
        raise e

class RepeatedRowError(Exception):
    """Raise for repeated rows."""


class ColumnDataTypeError(Exception):
    """Raise for columns with incorrect data types."""


class MissingColumnError(Exception):
    """Raise for required column missing."""


class UnexpectedColumn(Exception):
    """Raise for unexpected column."""


class ColumnValues(Exception):
    """Raise for unallowed column values."""


class ComponentError(Exception):
    """Raise for component class errors"""